use alloc::vec;
use alloc::vec::Vec;
use vapfs::{BlockDevice, Index};
use crate::read_datablock;
use crate::structs::{ListBlock, Superblock};

pub struct ListblockIter<'a> {
    listblock: &'a ListBlock,
    index: usize,
    buf1: (Index, Vec<u8>),
    buf2: (Index, Vec<u8>),
    buf3: (Index, Vec<u8>),
    buf4: (Index, Vec<u8>),
    buf5: (Index, Vec<u8>),
    buf6: (Index, Vec<u8>),
}

impl<'a> ListblockIter<'a> {
    pub fn new(listblock: &'a ListBlock) -> Self {
        Self { listblock, index: 0, buf1: (0, vec![]), buf2: (0, vec![]), buf3: (0, vec![]), buf4: (0, vec![]), buf5: (0, vec![]), buf6: (0, vec![]) }
    }

    pub fn next(&mut self, sb: &Superblock, bd: &mut dyn BlockDevice) -> Option<Index> {
        if self.index as Index >= self.listblock.count {
            return None;
        }

        if self.index < 32 {
            let index = self.listblock.direct_block_addresses[self.index];
            self.index += 1;
            return Some(index);
        } else {
            // copy pasted from structs.rs
            // if greater than 32, see the following:
            // let N = (maximum number of pointers in an indirect block) * 32
            // (beginning...count)
            // 32..N: single indirect block
            // 32+N..N^2: double indirect block
            // 32+N^2..N^3: triple indirect block
            // 32+N^3..N^4: quadruple indirect block
            // 32+N^4..N^5: quintuple indirect block
            // 32+N^5..N^6: sextuple indirect block

            // block index is address / (max_per_block ^ (1 if single, 2 if double, etc))
            // after getting your layer's address, repeat the previous layer's process

            let max_per_block = sb.block_size as u64 / 8;
            let N = max_per_block * 32;
            let N2 = N * N;
            let N3 = N2 * N;
            let N4 = N3 * N;
            let N5 = N4 * N;
            let N6 = N5 * N;

            let mut address = self.index as u64 - 32;

            // todo: you could probably rewrite this using like recursion or smth

            match address {
                _ if address < N => {
                    let block_index = address / max_per_block;
                    let address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != self.listblock.single_indirect_block_address[block_index as usize] {
                        self.buf1.1 = read_datablock(self.listblock.single_indirect_block_address[block_index as usize], sb, bd);
                        self.buf1.0 = self.listblock.single_indirect_block_address[block_index as usize];
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(address * 8) as usize..(address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N2 => {
                    let address = address - N;
                    let block_index = (address) / (max_per_block ^ 2);
                    if self.buf2.1.len() == 0 || self.buf2.0 != self.listblock.double_indirect_block_address[block_index as usize] {
                        self.buf2.1 = read_datablock(self.listblock.double_indirect_block_address[block_index as usize], sb, bd);
                        self.buf2.0 = self.listblock.double_indirect_block_address[block_index as usize];
                    }
                    let layer2_address = address / max_per_block;
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N3 => {
                    let address = address - N2;
                    let block_index = (address) / (max_per_block ^ 3);
                    if self.buf3.1.len() == 0 || self.buf3.0 != self.listblock.triple_indirect_block_address[block_index as usize] {
                        self.buf3.1 = read_datablock(self.listblock.triple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf3.0 = self.listblock.triple_indirect_block_address[block_index as usize];
                    }
                    let layer3_address = address / (max_per_block ^ 2);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N4 => {
                    let address = address - N3;
                    let block_index = (address) / (max_per_block ^ 4);
                    if self.buf4.1.len() == 0 || self.buf4.0 != self.listblock.quadruple_indirect_block_address[block_index as usize] {
                        self.buf4.1 = read_datablock(self.listblock.quadruple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf4.0 = self.listblock.quadruple_indirect_block_address[block_index as usize];
                    }
                    let layer4_address = address / (max_per_block ^ 3);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N5 => {
                    let address = address - N4;
                    let block_index = (address) / (max_per_block ^ 5);
                    if self.buf5.1.len() == 0 || self.buf5.0 != self.listblock.quintuple_indirect_block_address[block_index as usize] {
                        self.buf5.1 = read_datablock(self.listblock.quintuple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf5.0 = self.listblock.quintuple_indirect_block_address[block_index as usize];
                    }
                    let layer5_address = address / (max_per_block ^ 4);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf5.1[(layer5_address * 8) as usize..(layer5_address * 8 + 8) as usize]);
                    let layer4_block = u64::from_be_bytes(buf);
                    let layer4_address = (address % (max_per_block ^ 4)) / (max_per_block ^ 3);
                    if self.buf4.1.len() == 0 || self.buf4.0 != layer4_block {
                        self.buf4.1 = read_datablock(layer4_block, sb, bd);
                        self.buf4.0 = layer4_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ if address < N6 => {
                    let address = address - N5;
                    let block_index = (address) / (max_per_block ^ 6);
                    if self.buf6.1.len() == 0 || self.buf6.0 != self.listblock.sextuple_indirect_block_address[block_index as usize] {
                        self.buf6.1 = read_datablock(self.listblock.sextuple_indirect_block_address[block_index as usize], sb, bd);
                        self.buf6.0 = self.listblock.sextuple_indirect_block_address[block_index as usize];
                    }
                    let layer6_address = address / (max_per_block ^ 5);
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf6.1[(layer6_address * 8) as usize..(layer6_address * 8 + 8) as usize]);
                    let layer5_block = u64::from_be_bytes(buf);
                    let layer5_address = (address % (max_per_block ^ 5)) / (max_per_block ^ 4);
                    if self.buf5.1.len() == 0 || self.buf5.0 != layer5_block {
                        self.buf5.1 = read_datablock(layer5_block, sb, bd);
                        self.buf5.0 = layer5_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf5.1[(layer5_address * 8) as usize..(layer5_address * 8 + 8) as usize]);
                    let layer4_block = u64::from_be_bytes(buf);
                    let layer4_address = (address % (max_per_block ^ 4)) / (max_per_block ^ 3);
                    if self.buf4.1.len() == 0 || self.buf4.0 != layer4_block {
                        self.buf4.1 = read_datablock(layer4_block, sb, bd);
                        self.buf4.0 = layer4_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf4.1[(layer4_address * 8) as usize..(layer4_address * 8 + 8) as usize]);
                    let layer3_block = u64::from_be_bytes(buf);
                    let layer3_address = (address % (max_per_block ^ 3)) / (max_per_block ^ 2);
                    if self.buf3.1.len() == 0 || self.buf3.0 != layer3_block {
                        self.buf3.1 = read_datablock(layer3_block, sb, bd);
                        self.buf3.0 = layer3_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf3.1[(layer3_address * 8) as usize..(layer3_address * 8 + 8) as usize]);
                    let layer2_block = u64::from_be_bytes(buf);
                    let layer2_address = (address % (max_per_block ^ 2)) / max_per_block;
                    if self.buf2.1.len() == 0 || self.buf2.0 != layer2_block {
                        self.buf2.1 = read_datablock(layer2_block, sb, bd);
                        self.buf2.0 = layer2_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf2.1[(layer2_address * 8) as usize..(layer2_address * 8 + 8) as usize]);
                    let layer1_block = u64::from_be_bytes(buf);
                    let layer1_address = address % max_per_block;
                    if self.buf1.1.len() == 0 || self.buf1.0 != layer1_block {
                        self.buf1.1 = read_datablock(layer1_block, sb, bd);
                        self.buf1.0 = layer1_block;
                    }
                    let mut buf: [u8; 8] = [0; 8];
                    buf.copy_from_slice(&self.buf1.1[(layer1_address * 8) as usize..(layer1_address * 8 + 8) as usize]);
                    let block = u64::from_be_bytes(buf);
                    self.index += 1;
                    Some(block)
                }
                _ => None,
            }
        }
    }
}