use alloc::vec;
use ondisk_btree::ToBytes;
use vapfs::{BlockDevice, Timestamp};
use crate::btree::Directory;
use crate::{bitmap, structs, write_datablock, write_inode, write_journal_entry, write_superblock};
use crate::structs::{Filetype, Inode, JournalBlockWrite, JournalEntry, JournalEntryContents, ListBlock, Superblock, UNIXMode};

pub enum FilesystemError {
    /// Returned when an error relating to a lack of storage space occurs
    NotEnoughStorageSpace,
    /// Returned during filesystem initialisation if the block size is too small
    BlockSizeTooSmall,
    /// Returned when an error occurs that was not expected
    UnexpectedError,
}

/// initialises a filesystem with the given details
/// block_size is in bytes, and must be at least 2048 bytes
/// bd_size is in blocks, and should not exceed the size of the device that `bd` represents
/// current_time is the current time, in seconds since the unix epoch
/// inode_count is the number of inodes to create, or None to use the default (1/8 of the blocks available)
/// journal_count is the number of journal blocks to create, or None to use the default (128 blocks)
pub fn init_filesystem(bd: &mut dyn BlockDevice, block_size: u32, bd_size: u64, current_time: Timestamp, inode_count: Option<u64>, journal_count: Option<u64>) -> Result<(), FilesystemError> {
    if block_size < 2048 {
        return Err(FilesystemError::BlockSizeTooSmall);
    }
    let mut blocks_available = bd_size as i64;
    // if we have less than 5 blocks (probably around the minimum size), return not enough storage space
    if blocks_available < 5 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // minus one for the superblock
    blocks_available -= 1;
    // subtract journal count from blocks available
    blocks_available -= journal_count.unwrap_or(128) as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // inode count is the number of blocks available divided by 8
    let inode_count = inode_count.unwrap_or(blocks_available as u64 / 8);
    // each inode takes up one block, subtract that from the blocks available
    blocks_available -= inode_count as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // divide inode count by 8 (rounded up) to get the size of the inode bitmap
    let inode_bitmap_size = (inode_count + 7) / 8;
    // divide the bitmap size by the block size to get the number of blocks needed for the inode bitmap
    let inode_bitmap_blocks = (inode_bitmap_size + (block_size as u64 - 1)) / block_size as u64;
    // subtract the number of blocks needed for the inode bitmap from the blocks available
    blocks_available -= inode_bitmap_blocks as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }
    // divide the number of blocks available by 8 to get size of the data bitmap
    let data_bitmap_blocks = (blocks_available as u64 + 7) / 8;
    // divide the data bitmap size by the block size to get the number of blocks needed for the data bitmap
    let data_bitmap_blocks = (data_bitmap_blocks + (block_size as u64 - 1)) / block_size as u64;
    // subtract the number of blocks needed for the data bitmap from the blocks available
    blocks_available -= data_bitmap_blocks as i64;
    if blocks_available < 0 {
        return Err(FilesystemError::NotEnoughStorageSpace);
    }

    // layout of disk is roughly
    // superblock
    // data block bitmap
    // inode block bitmap
    // inode blocks
    // journal blocks
    // data blocks
    let sb_block = 1;
    let data_bitmap_block = sb_block + 1;
    let inode_bitmap_block = data_bitmap_block + data_bitmap_blocks;
    let inode_blocks = inode_bitmap_block + inode_bitmap_blocks;
    let journal_blocks = inode_blocks + inode_count;
    let data_blocks = journal_blocks + journal_count.unwrap_or(128);

    // create the superblock
    let mut sb = Superblock {
        magic: structs::MAGIC,
        block_size,
        first_data_block: data_blocks,
        first_inode_block: inode_blocks,
        first_journal_block: journal_blocks,
        inode_count,
        data_block_count: blocks_available as u64,
        journal_block_count: journal_count.unwrap_or(128),
        creation_time: current_time,
        last_modification_time: current_time,
        checksum: 0,
        journal_position: 0,
        reserved: [0; 7],
    };

    // calculate the checksum
    sb.convert_native_to_big_endian();
    sb.recalculate_checksum();
    sb.convert_big_endian_to_native();

    // create the first inode, which is the root directory
    let root_perms = UNIXMode::ReadOwner as u16 | UNIXMode::WriteOwner as u16 |
        UNIXMode::ExecuteOwner as u16 | UNIXMode::ReadGroup as u16 |
        UNIXMode::ExecuteGroup as u16 | UNIXMode::Read as u16 | UNIXMode::Execute as u16 |
        Filetype::Directory as u16;
    let mut root_inode = Inode {
        mode: root_perms,
        link_count: 0,
        uid: 0,
        gid: 0,
        size: 0,
        block_count: 0,
        creation_time: current_time,
        last_access_time: current_time,
        last_modification_time: current_time,
        last_inode_modification_time: current_time,
        deletion_time: 0,
        flags: 0,
        listblock: ListBlock {
            count: 0,
            direct_block_addresses: [0; 32],
            single_indirect_block_address: [0; 32],
            double_indirect_block_address: [0; 32],
            triple_indirect_block_address: [0; 32],
            quadruple_indirect_block_address: [0; 32],
            quintuple_indirect_block_address: [0; 32],
            sextuple_indirect_block_address: [0; 32],
        },
        checksum: 0,
    };

    let root_data = Directory::new();
    let root_bytes = root_data.to_bytes();
    // how many blocks are needed to store the root directory
    let root_block_count = (root_bytes.len() + (block_size as usize - 1)) / block_size as usize;
    // if its over 32 blocks, return unexpected error
    // (unless my b-tree implementation is awful, an empty directory should not take up 64KB)
    if root_block_count > 32 {
        return Err(FilesystemError::UnexpectedError);
    }
    let mut left = root_block_count;
    let mut buffer = vec![0; block_size as usize];
    let mut blocki = 0;
    let mut bufi = 0;
    let mut datai = 0;
    while left > 0 {
        buffer.clear();
        while bufi < block_size && datai < root_bytes.len() {
            buffer.push(root_bytes[datai]);
            bufi += 1;
            datai += 1;
        }
        unsafe { write_datablock(blocki, &sb, bd, &buffer) };
        blocki += 1;
        bufi = 0;
        left -= 1;
    }
    root_inode.size = root_bytes.len() as u64;
    root_inode.block_count = root_block_count as u64;
    for i in 0..root_block_count {
        root_inode.listblock.direct_block_addresses[i] = i as u64;
    }
    root_inode.listblock.count = root_block_count as u64;

    // recalculate the inode checksum
    root_inode.convert_native_to_big_endian();
    root_inode.recalculate_checksum();
    root_inode.convert_big_endian_to_native();

    // write the root inode to the first inode block
    unsafe {
        write_inode(0, &sb, bd, root_inode);
    }

    // create the inode bitmap
    let mut inode_bitmap = vec![0u8; inode_bitmap_size as usize];
    // set the first bit to 1, as the root inode is allocated
    inode_bitmap[0] |= 1;
    // write the inode bitmap to the inode bitmap block
    bd.seek(inode_bitmap_block * block_size as u64);
    bd.write_blocks(&inode_bitmap);

    // create the data bitmap
    let mut data_bitmap = vec![0u8; data_bitmap_blocks as usize];
    for i in 0..root_block_count {
        bitmap::set_bit(&mut data_bitmap, i, true);
    }
    // write the data bitmap to the data bitmap block
    bd.seek(data_bitmap_block * block_size as u64);
    bd.write_blocks(&data_bitmap);

    // zero out the journal blocks
    for i in 0..journal_blocks {
        unsafe {
            write_journal_entry(i, &sb, bd, JournalEntry {
                operation: 0,
                zeroed_content_crc32: 0,
                content: JournalEntryContents {
                    block_write: JournalBlockWrite {
                        flags: 0,
                        target_type: 0,
                        target_inode: 0,
                        target_block: 0,
                        real_target_block: 0,
                        source_block: 0,
                        source_block_crc32: 0,
                    },
                },
            });
        }
    }

    // write the superblock to the superblock block
    unsafe {
        write_superblock(sb, bd);
    }

    Ok(())
}